"""Support Mongo DB."""
from types import FunctionType

import bson
import marshmallow as ma
from flask._compat import string_types

from .filters import Filter as VanilaFilter, Filters
from .resource import ResourceOptions, Resource, APIError, logger


class ObjectId(ma.fields.Field):

    """ObjectID Marshmallow Field."""

    def _deserialize(self, value, attr, data):
        try:
            return bson.ObjectId(value)
        except:
            raise ma.ValidationError('invalid ObjectId `%s`' % value)

    def _serialize(self, value, attr, obj):
        if value is None:
            return ma.missing
        return str(value)


class MongoSchema(ma.Schema):

    """Serialize/deserialize results from mongo."""

    _id = ObjectId()

    def __init__(self, instance=None, **kwargs):
        """Initialize the schema."""
        self.instance = instance
        super(MongoSchema, self).__init__(**kwargs)

    @ma.post_load
    def make_instance(self, data):
        """Build object from data."""
        if self.instance is not None:
            self.instance.update(data)
            return self.instance

        return data

    def load(self, data, instance=None, *args, **kwargs):
        """Load data."""
        self.instance = instance or self.instance
        return super(MongoSchema, self).load(data, *args, **kwargs)


class MongoOptions(ResourceOptions):

    """Tune Mongo resource."""

    def __init__(self, cls):
        """Initialize the resource."""
        self._collection = None
        super(MongoOptions, self).__init__(cls)
        self.name = self.meta and getattr(self.meta, 'name', None)
        if not self.collection:
            return

        self.name = self.name or str(self.collection.name)

        if not cls.Schema:
            meta = type('Meta', (object,), self.schema_meta)
            cls.Schema = type(
                self.name.title() + 'Schema', (MongoSchema,), dict({'Meta': meta}, **self.schema))

    @property
    def collection(self):
        """Support lambdas as collection."""
        if isinstance(self._collection, FunctionType):
            return self._collection()
        return self._collection

    @collection.setter
    def collection(self, value):
        """Store initial values."""
        self._collection = value


class Filter(VanilaFilter):

    """Filter Mongo collections."""

    operators = {
        '$eq': '$eq',
        '$ge': '$gte',
        '$gt': '$gt',
        '$in': '$in',
        '$le': '$lte',
        '$lt': '$lt',
        '$ne': '$ne',
        '$nin': '$nin',
    }

    def apply(self, collection, ops, **kwargs):
        """Filter mongo."""
        logger.debug('Apply filter %s (%r)', self.name, ops)
        return collection.find({self.name: dict(ops)})


class MongoFilters(Filters):

    """Converter for Mongo resources."""

    FILTER_CLASS = Filter


class MongoChain(object):

    """Support query chains.

    Only for `find` and `find_one` methods.

    ::

        collection = MongoChain(mongo_collection)
        collection = collection.find({'field': 'value').find('field2': 'value')

        result = collection.find_one({'field3': 'value')
        results = collection.skip(10).limit(10)

    """

    CURSOR_METHODS = (
        'where', 'sort', 'skip', 'rewind', 'retrieved', 'remove_option', 'next', 'min',
        'max_time_ms', 'max_scan', 'max_await_time_ms', 'max', 'limit', 'hint', 'explain',
        'distinct', 'cursor_id', 'count', 'comment', 'collection', 'close', 'clone', 'batch_size',
        'alive', 'address', 'add_option', '__getitem__'
    )

    def __init__(self, collection):
        """Initialize the resource."""
        self.collection = collection
        self.query = []
        self.projection = None
        self.sorting = None

    def find(self, query=None, projection=None):
        """Store filters in self."""
        self.query = self.__update__(query)
        self.projection = projection
        return self

    def find_one(self, query=None, projection=None):
        """Apply filters and return cursor."""
        query = self.__update__(query)
        query = query and {'$and': query} or {}
        logger.debug('Mongo find one: %r', query)
        return self.collection.find_one(query, projection=projection)

    def aggregate(self, pipeline, **kwargs):
        """Aggregate collection."""
        if self.query:
            for params in pipeline:
                if '$match' in params:
                    query = self.__update__(params['$match'])
                    params['$match'] = {'$and': query}
                    break
            else:
                pipeline.insert(0, {'$match': {'$and': query}})
            logger.debug('Mongo aggregate: %r', pipeline)

        if self.sorting:
            pipeline = [p for p in pipeline if '$sort' not in p]
            pipeline.append({'$sort': dict(self.sorting)})

        return self.collection.aggregate(pipeline, **kwargs)

    def sort(self, key, direction=1):
        """Save ordering properties."""
        if isinstance(key, string_types):
            self.sorting = [(key, direction)]
        else:
            self.sorting = key

        return self

    def __repr__(self):
        """String representation."""
        return "<MongoChain (%s) %r>" % (self.collection.name, self.query)

    def __update__(self, query):
        """Update stored query."""
        if query:
            self.query.append(query)

        return self.query

    def __iter__(self):
        """Iterate by self collection."""
        query = self.query and {'$and': self.query} or {}
        if self.sorting:
            return self.collection.find(query, self.projection).sort(self.sorting)

        return self.collection.find(query, self.projection)

    def __getattr__(self, name):
        """Proxy any attributes except find to self.collection."""
        logger.debug('Mongo load: %r', self.query)
        if name in self.CURSOR_METHODS:
            query = self.query and {'$and': self.query} or {}
            cursor = self.collection.find(query, self.projection)
            if self.sorting:
                cursor = cursor.sort(self.sorting)
            return getattr(cursor, name)
        return getattr(self.collection, name)


class MongoResource(Resource):

    """Provide API for Pymongo document and collections."""

    OPTIONS_CLASS = MongoOptions

    class Meta:

        """Default params."""

        collection = None
        filters = 'login',
        filters_converter = MongoFilters
        aggregate = False  # Support aggregation. Set to pipeline.
        object_id = '_id'
        schema = {}

    def get_many(self, *args, **kwargs):
        """Return collection filters."""
        return MongoChain(self.meta.collection)

    def get_one(self, *args, **kwargs):
        """Load a resource."""
        resource = super(MongoResource, self).get_one(*args, **kwargs)
        if not resource:
            return None

        return self.collection.find_one({self.meta.object_id: bson.ObjectId(resource)})

    def paginate(self, offset=0, limit=None):
        """Paginate collection."""
        if self.meta.aggregate:
            pipeline_all = self.meta.aggregate + [{'$skip': offset}, {'$limit': limit}]
            pipeline_num = self.meta.aggregate + [{'$group': {'_id': None, 'total': {'$sum': 1}}}]
            counts = list(self.collection.aggregate(pipeline_num))
            return (
                self.collection.aggregate(pipeline_all),
                counts and counts[0]['total'] or 0
            )
        return self.collection.skip(offset).limit(limit), self.collection.count()

    def to_simple(self, data, many=False, **kwargs):
        """Support aggregation."""
        if isinstance(data, MongoChain) and self.meta.aggregate:
            data = data.aggregate(self.meta.aggregate)
        return super(MongoResource, self).to_simple(data, many=many, **kwargs)

    def get_schema(self, resource=None, **kwargs):
        """Create the resource schema."""
        return self.Schema(instance=resource)  # noqa

    def save(self, resource):
        """Save resource to DB."""
        if resource.get('_id'):
            self.meta.collection.replace_one({'_id': resource['_id']}, resource)
        else:
            write = self.meta.collection.insert_one(resource)
            resource['_id'] = write.inserted_id
        return resource

    def sort(self, collection, *sorting, **Kwargs):
        """Sort resources."""
        sorting = {name: -1 if desc else 1 for name, desc in sorting}
        return collection.sort(list(sorting.items()))

    def delete(self, resource=None, **kwargs):
        """Delete a resource from Mongo collection."""
        if resource is None:
            raise APIError('Resource not found', status_code=404)
        self.collection.delete_one({self.meta.object_id: resource[self.meta.object_id]})



"""Support Peewee ORM."""
from __future__ import absolute_import
from peewee import SQL, Field
from flask._compat import string_types

from .resource import ResourceOptions, Resource, APIError, logger
from .filters import Filter as VanilaFilter, Filters

try:
    from marshmallow_peewee import ModelSchema
except ImportError:
    import logging
    logging.error('Marshmallow-Peewee should be installed to use the integration.')
    raise


def ensure_join(qs, lm, rm, on=None, **join_kwargs):
    """TODO: remove me when problem in Peewee will be fixed."""
    ctx = qs._query_ctx
    for join in qs._joins.get(lm, []):
        if join.dest is rm:
            return qs
    return qs.switch(lm).join(rm, on=on, **join_kwargs).switch(ctx)


class Filter(VanilaFilter):

    """Filter Peewee Collection."""

    operators = VanilaFilter.operators
    operators['$in'] = lambda v, c: v << c
    operators['$nin'] = lambda v, c: ~(v << c)
    operators['$none'] = lambda f, v: f >> v
    operators['$like'] = lambda f, v: f % v
    operators['$ilike'] = lambda f, v: f ** v
    operators['$contains'] = lambda f, v: f.contains(v)
    operators['$starts'] = lambda f, v: f.startswith(v)
    operators['$ends'] = lambda f, v: f.endswith(v)
    operators['$between'] = lambda f, v: f.between(*v)
    operators['$regexp'] = lambda f, v: f.regexp(v)

    list_ops = VanilaFilter.list_ops + ('$between', '$nin')

    mfield = None

    def __init__(self, name, fname=None, field=None, mfield=None):
        """Initialize filter."""
        super(Filter, self).__init__(name, fname, field)
        self.mfield = mfield or self.mfield

    def apply(self, collection, ops, view=None, **kwargs):
        """Filter given Peewee collection."""
        if not self.mfield and view is None:
            return collection

        logger.debug('Apply filter %s (%r)', self.name, ops)

        # Auto join to another collection
        if self.mfield and hasattr(self.mfield, 'model_class') and \
                self.mfield.model_class is not view.meta.model:
            collection = ensure_join(collection, view.meta.model, self.mfield.model_class)

        mfield = self.mfield or view.meta.model._meta.fields.get(self.field.attribute)
        return collection.where(*[op(mfield, val) for op, val in ops])


class ModelFilters(Filters):

    """Filter Peewee Collection."""

    FILTER_CLASS = Filter


class ModelResourceOptions(ResourceOptions):

    """Peewee Resource Options."""

    def __init__(self, cls):
        """Get meta from given model."""
        super(ModelResourceOptions, self).__init__(cls)
        self.name = (self.meta and getattr(self.meta, 'name', None)) or \
            self.model and self.model._meta.db_table or self.name

        self.sorting = dict(
            (isinstance(n, Field) and n.name or n, prop)
            for (n, prop) in self.sorting.items())

        if not self.model:
            return None

        if not cls.Schema:
            meta = type('Meta', (object,), dict({'model': self.model}, **self.schema_meta))
            if self.models_converter:
                meta.model_converter = self.models_converter

            cls.Schema = type(
                self.name.title() + 'Schema', (ModelSchema,), dict({'Meta': meta}, **self.schema))


class ModelResource(Resource):

    """Provide API for Peewee models."""

    OPTIONS_CLASS = ModelResourceOptions

    class Meta:

        """Default options."""

        model = None
        filters_converter = ModelFilters
        models_converter = None
        schema = {}

    def get_many(self, *args, **kwargs):
        """Setup queryset."""
        return self.meta.model.select()

    def sort(self, collection, *sorting, **Kwargs):
        """Sort resources."""
        logger.debug('Sort collection: %r', sorting)
        sorting_ = []
        for field, desc in sorting:
            if isinstance(field, string_types):
                field = self.meta.model._meta.fields.get(field) or SQL(field)

            if desc:
                field = field.desc()
            sorting_.append(field)
        if sorting_:
            collection = collection.order_by(*sorting_)
        return collection

    def get_one(self, *args, **kwargs):
        """Load a resource."""
        resource = super(ModelResource, self).get_one(*args, **kwargs)
        if not resource:
            return None

        try:
            resource = self.collection.where(self.meta.model._meta.primary_key == resource).get()
        except self.meta.model.DoesNotExist:
            raise APIError('Resource not found', status_code=404)

        return resource

    def get_schema(self, resource=None, **kwargs):
        """Put resource to schema."""
        return self.Schema(instance=resource)

    def save(self, resource):
        """Save resource to DB."""
        resource.save()
        return resource

    def delete(self, resource=None, **kwargs):
        """Delete a resource."""
        if resource is None:
            raise APIError('Resource not found', status_code=404)
        resource.delete_instance()

    def paginate(self, offset=0, limit=None):
        """Paginate queryset."""
        logger.debug('Paginate collection, offset: %d, limit: %d', offset, limit)
        qs = self.collection.order_by()
        if qs._group_by:
            qs._select = qs._group_by

        return self.collection.offset(offset).limit(limit), qs.count()

# pylama:ignore=E1102,W0212



from __future__ import absolute_import

from types import FunctionType
from sqlalchemy import func
from sqlalchemy.orm.attributes import QueryableAttribute
from flask._compat import string_types

from .filters import Filter as VanilaFilter, Filters
from .resource import ResourceOptions, Resource, APIError, logger


try:
    from sqlalchemy.inspection import inspect
    from marshmallow_sqlalchemy import ModelSchema
except ImportError:
    logger.error('Marshmallow-SQLAlchemy should be installed to use the integration.')
    raise


class Filter(VanilaFilter):

    operators = VanilaFilter.operators
    operators['$in'] = lambda c, v: c.in_(v)
    operators['$nin'] = lambda c, v: ~c.in_(v)

    mfield = None

    def __init__(self, name, fname=None, field=None, mfield=None):
        """Initialize filter.

        :param mfield: Model field
        """
        super(Filter, self).__init__(name, fname, field)
        self.mfield = mfield if mfield is not None else self.mfield

    def apply(self, collection, ops, view=None, **kwargs):
        if self.mfield is not None and view is None:
            return collection

        logger.debug('Apply filter %s (%r)', self.name, ops)

        mfield = self.mfield if self.mfield is not None else getattr(view.meta.model, self.name, None)
        if mfield is None:
            return collection
        collection = collection.filter(*(op(mfield, val) for op, val in ops))
        return collection


class ModelFilters(Filters):

    FILTER_CLASS = Filter


class ModelResourceOptions(ResourceOptions):

    def __init__(self, cls):
        self._session = None

        super(ModelResourceOptions, self).__init__(cls)

        self.sorting = dict(
            (isinstance(n, QueryableAttribute) and n.name or n, prop)
            for (n, prop) in self.sorting.items())

        self.name = (self.meta and getattr(self.meta, 'name', None)) or \
            self.model and self.model.__tablename__ or self.name

        if not self.model:
            return None

        if not cls.Schema:
            meta = type('Meta', (object,), dict({'model': self.model}, **self.schema_meta))
            cls.Schema = type(
                self.name.title() + 'Schema', (ModelSchema,), dict({'Meta': meta}, **self.schema))

        if not self.primary_key:
            self.primary_key = inspect(self.model).primary_key[0]

        # Flask-SQLAlchemy support
        if not self.session and hasattr(self.model, 'query'):
            self.session = self.model.query.session

    @property
    def session(self):
        """Support lambdas as session."""
        if isinstance(self._session, FunctionType):
            return self._session()
        return self._session

    @session.setter
    def session(self, value):
        """Store initial values."""
        self._session = value


class ModelResource(Resource):

    """Provide API for SQLAlchemy models."""

    OPTIONS_CLASS = ModelResourceOptions

    class Meta:
        filters_converter = ModelFilters
        model = None
        primary_key = None
        schema = {}
        session = None

    def get_many(self, *args, **kwargs):
        return self.meta.session.query(self.meta.model).filter()

    def sort(self, collection, *sorting, **kwargs):
        sorting_ = []
        for prop, desc in sorting:
            if isinstance(prop, string_types):
                prop = getattr(self.meta.model, prop, None)

            if prop is None:
                continue

            if desc:
                prop = prop.desc()

            sorting_.append(prop)

        if sorting_:
            collection = collection.order_by(None).order_by(*sorting_)
        return collection

    def get_one(self, *args, **kwargs):
        """Load a resource."""
        resource = super(ModelResource, self).get_one(*args, **kwargs)
        if not resource:
            return None

        resource = self.collection.filter(self.meta.primary_key == resource).first()
        if resource is None:
            raise APIError('Resource not found', status_code=404)

        return resource

    def get_schema(self, resource=None, **kwargs):
        return self.Schema(session=self.meta.session, instance=resource)

    def save(self, resource):
        """Save resource to DB."""
        self.meta.session.add(resource)
        self.meta.session.commit()
        return resource

    def delete(self, resource=None, **kwargs):
        """Delete a resource."""
        if resource is None:
            raise APIError('Resource not found', status_code=404)
        self.meta.session.delete(resource)
        self.meta.session.commit()

    def paginate(self, offset=0, limit=None):
        """Paginate queryset."""
        cqs = self.collection.with_entities(func.count()).order_by(None)
        return self.collection.offset(offset).limit(limit), self.meta.session.execute(cqs).scalar()



import datetime as dt

import marshmallow as ma


class Timestamp(ma.fields.Field):

    default_error_messages = {
        'invalid': 'Not a valid timestamp.'
    }

    def _serialize(self, value, attr, obj):
        """Serialize given datetime to timestamp."""
        if value is None:
            return None

        return int(datetime_to_timestamp(value))

    def _deserialize(self, value, attr, data):
        if not value:  # Falsy values, e.g. '', None, [] are not valid
            raise self.fail('invalid')

        try:
            return dt.datetime.utcfromtimestamp(float(value))
        except ValueError:
            raise self.fail('invalid')


class MSTimestamp(Timestamp):

    def _serialize(self, value, *args):
        """Serialize given datetime to timestamp."""
        if value is not None:
            value = super(MSTimestamp, self)._serialize(value, *args) * 1e3
        return value

    def _deserialize(self, value, *args):
        if value:
            value = int(value) / 1e3

        return super(MSTimestamp, self)._deserialize(value, *args)


if ma.compat.PY2:
    def datetime_to_timestamp(dt_, epoch=dt.datetime(1970, 1, 1)):
        """Convert given datetime object to timestamp in seconds."""
        return (dt_.replace(tzinfo=None) - epoch).total_seconds()

else:

    def datetime_to_timestamp(dt_):
        """Convert given datetime object to timestamp in seconds."""
        return dt_.replace(tzinfo=dt.timezone.utc).timestamp()



"""Support filters."""

import json
import operator

from cached_property import cached_property
from flask import request
from marshmallow import fields, missing, ValidationError
from . import logger


FILTERS_ARG = 'where'


class Filter(object):
    """Base filter class."""

    operators = {
        '$lt': operator.lt,
        '$le': operator.le,
        '$gt': operator.gt,
        '$ge': operator.ge,
        '$eq': operator.eq,
        '$ne': operator.ne,
        '$in': lambda v, c: v in c,
    }

    list_ops = '$in',

    def __init__(self, name, fname=None, field=None):
        """Initialize filter."""
        self.name = name
        self.fname = fname or name
        self.field = field or fields.Raw(attribute=name)

    def __repr__(self):
        return '<Filter %s>' % (self.field.attribute or self.name or self.fname)

    def parse(self, data):
        """Parse operator and value from filter's data."""
        val = data.get(self.fname, missing)
        if not isinstance(val, dict):
            val = self.field.deserialize(val)
            request.filters[self.fname] = val
            return (self.operators['$eq'], val),

        ops = ()
        request.filters[self.fname] = {}
        for op, val in val.items():
            if op not in self.operators:
                continue
            val = self.field.deserialize(val) if op not in self.list_ops else [self.field.deserialize(v) for v in val]  # noqa
            ops += (self.operators[op], val),
            request.filters[self.fname][op] = val

        return ops

    def filter(self, collection, data, **kwargs):
        """Parse data and apply filter."""
        try:
            return self.apply(collection, self.parse(data), **kwargs)
        except ValidationError:
            return collection

    def apply(self, collection, ops, **kwargs):  # noqa
        """Apply current filter."""
        def validator(obj):
            return all(op(obj, val) for (op, val) in ops)
        return [o for o in collection if validator(o)]


class Filters(object):
    """Filters helper."""

    FILTER_CLASS = Filter

    def __init__(self, filters, View):
        """Initialize the helper."""
        self._filters = filters
        self.View = View

    @cached_property
    def filters(self):
        """Prepare filters."""
        if not self._filters:
            return None
        return list(f if isinstance(f, Filter) else self.convert(f) for f in self._filters)

    def convert(self, name):
        """Setup a filter by name."""
        if not self.View.Schema or name not in self.View.Schema._declared_fields:  # noqa
            return self.FILTER_CLASS(name)
        field = self.View.Schema._declared_fields[name]  # noqa
        return self.FILTER_CLASS(name, field=field)

    def filter(self, collection, view, *args, **kwargs):
        """Filter the given collection."""
        request.filters = {}
        data = request.args.get(FILTERS_ARG)
        if not data or self.filters is None:
            return collection

        try:
            data = json.loads(data)
        except (ValueError, TypeError):
            return collection

        logger.debug('Filter resources: %r', data)

        filters = [f for f in self.filters if f.fname in data]
        logger.debug('Filters active: %r', filters)
        for f in filters:
            collection = f.filter(collection, data, view=view, **kwargs)
        return collection



from __future__ import absolute_import

from flask import Blueprint, jsonify, request, render_template, json
from flask._compat import string_types, PY2
import os
import urllib
import warnings
from inspect import isclass

from . import APIError
from .auth import current_user

from .resource import Resource
from apispec.ext.marshmallow.swagger import schema2jsonschema

if PY2:
    urlencode = urllib.urlencode
else:
    urlencode = urllib.parse.urlencode


DEFAULT = object()
STATIC = os.path.abspath(os.path.join(os.path.dirname(__file__), 'static'))
TEMPLATE = os.path.abspath(os.path.join(os.path.dirname(__file__), 'templates'))


class Api(Blueprint):

    """Implement REST API."""

    def __init__(self, name, import_name, specs=True, version="1", url_prefix=None, **kwargs):
        self.version = version
        self.specs = specs

        if not url_prefix and version:
            url_prefix = "/%s" % version

        if self.specs:
            kwargs['static_folder'] = STATIC
            kwargs['template_folder'] = TEMPLATE

        super(Api, self).__init__(name, import_name, url_prefix=url_prefix, **kwargs)
        self.app = None
        self.resources = []

    def register(self, app, options=None, first_registration=False):
        """Register self to application."""
        self.app = app
        app.errorhandler(APIError)(self.handle_error)
        if self.specs:
            self.route('/_specs')(self.specs_view)

            @self.route('/')
            def specs_html(): # noqa
                return render_template('swagger.html')

        return super(Api, self).register(app, options or {}, first_registration)

    def authorize(self, *args, **kwargs):
        """Make authorization process.

        The logic could be redifined for each resource.
        """
        return current_user

    def authorization(self, callback):
        """A decorator which helps to update authorize method for current Api."""
        self.authorize = callback
        return callback

    @staticmethod
    def handle_error(error):
        response = jsonify(error.to_dict())
        response.status_code = error.status_code
        return response

    def connect(self, *args, **kwargs):
        warnings.warn('The @connect method is depricated, use @route instead.')
        return self.route(*args, **kwargs)

    def route(self, resource=None, url=None, url_detail=DEFAULT, **options):
        """Connect resource to the API."""

        api = self

        def wrapper(res):

            if not isclass(res):
                res = Resource.from_func(res, **options)

            elif not issubclass(res, Resource):
                raise ValueError('Resource should be subclass of api.Resource.')

            api.resources.append(res)

            url_ = res.meta.url = url or res.meta.url or ('/%s' % res.meta.name)
            view_func = res.as_view(res.meta.name, api)
            api.add_url_rule(url_, view_func=view_func, **options)

            for _, (route_, endpoint_, options_) in res.meta.endpoints.values():
                api.add_url_rule('%s/%s' % (url_, route_.strip('/')), view_func=view_func,
                                 defaults={'endpoint': endpoint_}, **options_)

            url_detail_ = url_detail
            if url_detail is DEFAULT:
                url_detail_ = res.meta.url_detail = res.meta.url_detail or \
                    ('%s/<%s>' % (url_, res.meta.name))

            if url_detail:
                api.add_url_rule(url_detail_, view_func=view_func, **options)

            if api.app is not None:
                Blueprint.register(api, api.app, {}, False)

            return res

        if resource is not None and isinstance(resource, type) and issubclass(resource, Resource):
            return wrapper(resource)

        elif isinstance(resource, string_types):
            url = resource

        return wrapper

    def run(self, Resource, path=None, query_string=None, kwargs=None, **rkwargs):
        """Run given resource manually.

        See werkzeug.test.EnvironBuilder for rkwargs.
        """
        resource = Resource(self, raw=True)

        if isinstance(query_string, dict):
            if 'where' in query_string:
                query_string['where'] = json.dumps(query_string['where'])
            query_string = urlencode(query_string)

        rkwargs['query_string'] = query_string
        args = []
        if path:
            args.append(path)
        ctx = self.app.test_request_context(*args, **rkwargs)

        with ctx:
            kwargs = kwargs or {}
            return resource.dispatch_request(**kwargs)

    def specs_view(self, *args, **kwargs):
        specs = {
            'openapi': '3.0.0',
            'info': {
                'title': self.name,
                'description': self.__doc__,
                'version': self.version,
            },
            'basePath': self.url_prefix,
            'tags': [],
            'paths': {},
            'components': {'schemas': {}},
            'host': request.host,
        }

        for resource in self.resources:

            defaults = {
                'consumes': ['application/json'],
                'produces': ['application/json'],
                'security': [{'api_key': []}],
                'tags': [resource.meta.name],
                'responses': {
                    200: {
                        'description': 'OK',
                        'content': {
                            'application/json': {
                            }
                        }
                    }
                }
            }
            if resource.Schema:
                jsonschema = schema2jsonschema(resource.Schema)
                for prop in jsonschema.get('properties', {}).values():
                    if prop.get('type') == 'string' and prop.get('default'):
                        prop['default'] = str(prop['default'])

                specs['components']['schemas'][resource.meta.name] = jsonschema
                defaults['responses'][200]['content']['application/json']['schema'] = {
                    '$ref':  '#/components/schemas/{}'.format(resource.meta.name)
                }

            specs['tags'].append({
                'name': resource.meta.name,
                'description': resource.__doc__ or resource.__class__.__doc__,
            })

            for endpoint, (url_, name_, params_) in resource.meta.endpoints.values():
                specs['paths'][
                    "%s/%s" % (resource.meta.url.rstrip('/'), url_flask_to_swagger(url_))] = path = {}
                path['get'] = dict(
                    summary=endpoint.__doc__,
                    description=endpoint.__doc__,
                    **defaults)
                if hasattr(endpoint, 'specs'):
                    path['get'].update(endpoint.specs)

            specs['paths'][resource.meta.url] = path = {}
            for method in ('get', 'post'):
                if method.upper() not in resource.methods or not hasattr(resource, 'post'):
                    continue
                view = getattr(resource, method)
                path[method] = dict(summary=view.__doc__, description=view.__doc__, **defaults)

                if method == 'post':
                    params = {
                        'in': 'body',
                        'name': 'body',
                        'description': 'resource body',
                        'required': True,
                        'schema': {},
                    }
                    if resource.Schema:
                        params['schema']['$ref'] = '#/components/schemas/{}'.format(
                            resource.meta.name)

                    path[method]['parameters'] = [params]

                if resource.meta.specs:
                    path[method].update(resource.meta.specs)

            if resource.meta.url_detail:
                url_detail = url_flask_to_swagger(resource.meta.url_detail)
                path = specs['paths'][url_detail] = {}
                for method in ('get', 'put', 'delete'):
                    if method.upper() not in resource.methods or not hasattr(resource, 'post'):
                        continue
                    view = getattr(resource, method)
                    path[method] = dict(
                        summary=view.__doc__, description=view.__doc__,
                        parameters=[{
                            'name': resource.meta.name,
                            'in': 'path',
                            'description': 'ID of resource',
                            'type': 'string',
                            'required': True
                        }], **defaults)

                    if method == 'put':
                        params = {
                            'in': 'body',
                            'name': 'body',
                            'description': 'resource body',
                            'required': True,
                            'schema': {},
                        }
                        if resource.Schema:
                            params['schema']['$ref'] = '#/components/schemas/{}'.format(
                                resource.meta.name)
                        path[method]['parameters'].append(params)

                if resource.meta.specs:
                    path[method].update(resource.meta.specs)

        if isinstance(self.specs, dict):
            specs.update(self.specs)

        return specs


def url_flask_to_swagger(source):
    """Convert Flask URL to swagger path."""
    return source.replace('<', '{').replace('>', '}')



import logging


__license__ = "MIT"
__project__ = "Flask-Restler"
__version__ = "1.17.1"


logger = logging.getLogger('flask-restler')
logger.addHandler(logging.NullHandler())


class APIError(Exception):

    """Store API exception's information."""

    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['error'] = self.message
        rv['code'] = rv.get('code', self.status_code)
        return rv


def route(rule=None, endpoint=None, **options):
    """Custom routes in resources."""

    def decorator(f):
        endpoint_ = endpoint or f.__name__.lower()
        f.route = (rule, endpoint_, options)
        return f

    if callable(rule):
        rule, f = rule.__name__.lower(), rule
        return decorator(f)

    return decorator

from .api import Api, Resource  # noqa



"""Base API resource."""

from __future__ import absolute_import

import collections
import math
import logging

from flask import request, current_app, abort, Response
from flask._compat import with_metaclass
from flask.json import dumps
from flask.views import View

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from . import APIError, logger
from .auth import current_user
from .filters import Filters, FILTERS_ARG


PER_PAGE_ARG = 'per_page'
PAGE_ARG = 'page'
SORT_ARG = 'sort'
INTERNAL_ARGS = set([PER_PAGE_ARG, PAGE_ARG, SORT_ARG, FILTERS_ARG])


class ResourceOptions(object):
    """Prepare resource options."""

    def __init__(self, cls):
        """Initialize resources' options."""
        # Store link to self.meta
        self.meta = meta = getattr(cls, "Meta", None)

        self.cls = cls

        # Inherit meta from parents
        for base in reversed(cls.mro()):
            if not hasattr(base, "Meta"):
                continue

            for k, v in base.Meta.__dict__.items():
                if k.startswith('__'):
                    continue
                setattr(self, k, v)

        # Generate name
        self.name = (meta and getattr(meta, 'name', None)) or \
            cls.__name__.lower().split('resource', 1)[0]

        if self.per_page:  # noqa
            self.per_page = int(self.per_page)

        if self.specs:  # noqa
            self.specs = dict(self.specs)

        if self.strict:  # noqa
            if not isinstance(self.strict, collections.Iterable):
                self.strict = INTERNAL_ARGS
            self.strict = set(self.strict) | INTERNAL_ARGS

        # Setup endpoints
        self.endpoints = getattr(self, 'endpoints', {})
        self.endpoints.update({
            value.route[1]: (value, value.route) for value in cls.__dict__.values()
            if hasattr(value, 'route') and isinstance(value.route, tuple)
        })

        # Setup schema_meta
        self.schema_meta = self.schema_meta or {
            k[7:]: self.__dict__[k] for k in self.__dict__
            if k.startswith('schema_') and not k == 'schema_meta'
        }

        # Setup filters
        self.filters = self.filters_converter(self.filters, cls)

        # Setup sorting
        self.sorting = dict(n if isinstance(n, (list, tuple)) else (n, n) for n in self.sorting)

    def __repr__(self):
        return "<Options %s>" % self.cls


class ResourceMeta(type):
    """Do some work for resources."""

    def __new__(mcs, name, bases, params):
        """Initialize class."""
        cls = super(ResourceMeta, mcs).__new__(mcs, name, bases, params)
        cls.methods = set([method.upper() for method in cls.methods])
        cls.meta = cls.OPTIONS_CLASS(cls)
        return cls


class Resource(with_metaclass(ResourceMeta, View)):
    """Base API Resource object."""

    OPTIONS_CLASS = ResourceOptions

    # methods: Allowed methods
    methods = 'get',

    # Schema: Resource marshmallow schema
    Schema = None

    class Meta:
        """Tune the resource."""

        # name: Resource's name (if it is None, it will be calculated)
        name = None

        # per_page: Paginate results (set to None for disable pagination)
        per_page = 100

        # link_header: Add Link header with pagination
        page_link_header = True

        # url: URL for collection, if it is None it will be calculated
        # url_detail: URL for resource detail, if it is None it will be calculated
        url = url_detail = None

        # Resource filters
        filters = ()

        # Define allowed resource sorting params
        sorting = ()

        # Filters converter class
        filters_converter = Filters

        # Strict mode (only allowed query params) set to list of names for allowed query params
        strict = False

        # Swagger specs
        specs = None

        # marshmallow.Schema.Meta options
        # -------------------------------

        # Redefine Schema.Meta completely
        schema_meta = None

    def __init__(self, api=None, raw=False, **kwargs):
        """Initialize the resource."""
        self.api = api
        self.raw = raw
        self.auth = self.collection = None
        super(Resource, self).__init__(**kwargs)

    @classmethod
    def from_func(cls, func, methods=None, **options):

        if methods is None:
            methods = ['GET']

        def proxy(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        params = {m.lower(): proxy for m in methods}
        params['methods'] = methods
        return type(func.__name__, (cls,), params)

    def dispatch_request(self, *args, **kwargs):
        """Process current request."""
        if self.meta.strict and not (self.meta.strict >= set(request.args)):
            raise APIError('Invalid query params.')

        self.auth = self.authorize(*args, **kwargs)
        self.collection = self.get_many(*args, **kwargs)

        kwargs['resource'] = resource = self.get_one(*args, **kwargs)

        endpoint = kwargs.pop('endpoint', None)
        if endpoint and hasattr(self, endpoint):
            method = getattr(self, endpoint)
            logger.debug('Loaded endpoint: %s', endpoint)
            response = method(*args, **kwargs)
            return self.to_json_response(response)

        headers = {}

        if request.method == 'GET' and resource is None:

            # Filter resources
            self.collection = self.filter(self.collection, *args, **kwargs)

            # Sort resources
            if SORT_ARG in request.args:
                sorting = ((name.strip('-'), name.startswith('-'))
                           for name in request.args[SORT_ARG].split(','))
                sorting = (
                    (self.meta.sorting.get(n), d) for n, d in sorting if n in self.meta.sorting)
                self.collection = self.sort(self.collection, *sorting, **kwargs)

            # Paginate resources
            if self.meta.per_page:
                try:
                    per_page = int(request.args.get(PER_PAGE_ARG, self.meta.per_page))
                    if per_page:
                        page = int(request.args.get(PAGE_ARG, 0))
                        offset = page * per_page
                        self.collection, total = self.paginate(offset, per_page)
                        headers = make_pagination_headers(
                            per_page, page, total, self.meta.page_link_header)
                except ValueError:
                    raise APIError('Pagination params are invalid.')

        if logger.level <= logging.DEBUG:
            logger.debug('Collection: %r', self.collection)
            logger.debug('Params: %r', kwargs)

        try:
            method = getattr(self, request.method.lower())
        except AttributeError:
            return abort(405)

        response = method(*args, **kwargs)
        return self.to_json_response(response, headers=headers)

    def to_json_response(self, response, headers=None):
        """Serialize simple response to Flask response."""
        if self.raw or isinstance(response, Response):
            return response
        response = current_app.response_class(
            dumps(response, indent=2), mimetype='application/json')
        if headers:
            response.headers.extend(headers)
        return response

    def authorize(self, *args, **kwargs):
        """Default authorization method."""
        if self.api is not None:
            return self.api.authorize(self, *args, **kwargs)
        return current_user

    def get_many(self, *args, **kwargs):
        """Get collection."""
        return []

    def get_one(self, *args, **kwargs):
        """Load resource."""
        return kwargs.get(self.meta.name)

    def get_schema(self, resource=None, **kwargs):
        """Get schema."""
        return self.Schema and self.Schema()  # noqa

    def filter(self, collection, *args, **kwargs):
        """Filter collection."""
        return self.meta.filters.filter(collection, self, *args, **kwargs)

    def sort(self, collection, *sorting, **kwargs):
        """Sort collection."""
        logger.debug('Sort collection: %r', sorting)
        return collection

    def load(self, data, resource=None, **kwargs):
        """Load given data into schema."""
        schema = self.get_schema(resource=resource, **kwargs)
        resource, errors = schema.load(data, partial=resource is not None)
        if errors:
            raise APIError('Bad request', payload={'errors': errors})
        return resource

    def save(self, resource):
        """Create a resource."""
        return resource

    def to_simple(self, data, many=False, **kwargs):
        """Serialize response to simple object (list, dict)."""
        schema = self.get_schema(many=many, **kwargs)
        return schema.dump(data, many=many).data if schema else data

    def paginate(self, offset, limit):
        """Paginate results."""
        logger.debug('Paginate collection, offset: %d, limit: %d', offset, limit)
        return self.collection[offset: offset + limit], len(self.collection)

    def get(self, resource=None, **kwargs):
        """Get resource or collection of resources."""
        logger.debug('Get resources (%r)', resource)
        if resource is not None and resource != '':
            return self.to_simple(resource, resource=resource, **kwargs)

        return self.to_simple(self.collection, many=True, **kwargs)

    def post(self, **kwargs):
        """Create a resource."""
        data = request.json or {}
        resource = self.load(data, **kwargs)
        resource = self.save(resource)
        logger.debug('Create a resource (%r)', kwargs)
        return self.to_simple(resource, **kwargs)

    def put(self, resource=None, **kwargs):
        """Update a resource."""
        logger.debug('Update a resource (%r)', resource)
        if resource is None:
            raise APIError('Resource not found', status_code=404)

        return self.post(resource=resource, **kwargs)

    patch = put

    def delete(self, resource=None, **kwargs):
        """Delete a resource."""
        logger.debug('Delete a resource (%r)', resource)
        if resource is None:
            raise APIError('Resource not found', status_code=404)
        self.collection.remove(resource)


def make_pagination_headers(limit, curpage, total, link_header=True):
    """Return Link Hypermedia Header."""
    lastpage = int(math.ceil(1.0 * total / limit) - 1)
    headers = {'X-Total-Count': str(total), 'X-Limit': str(limit),
               'X-Page-Last': str(lastpage), 'X-Page': str(curpage)}

    if not link_header:
        return headers

    base = "{}?%s".format(request.path)
    links = {}
    links['first'] = base % urlencode(dict(request.args, **{PAGE_ARG: 0}))
    links['last'] = base % urlencode(dict(request.args, **{PAGE_ARG: lastpage}))
    if curpage:
        links['prev'] = base % urlencode(dict(request.args, **{PAGE_ARG: curpage - 1}))
    if curpage < lastpage:
        links['next'] = base % urlencode(dict(request.args, **{PAGE_ARG: curpage + 1}))

    headers['Link'] = ",".join(['<%s>; rel="%s"' % (v, n) for n, v in links.items()])
    return headers


# pylama:ignore=R0201
